/*-
|> File
|| drawcmd.c
||
|> Description
||
|> Comments
||
|> History
|| 950jj _mf Created.
||
||==================================================================|>
*/
#ifndef _NOIDENT
#ident "$Id: drawcmd.c,v 2.8 1995/11/23 17:00:13 mreynold Exp $"
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "printer.h"
#include "printlib.h"
#include "drawcmd.h"
#include "global.h"

extern struct PrintClient PCS[];

/* current GC values pointer */
static XGCValues       *currGcvp = &initial_GC;

/* font information mapping tables and cache of loaded fonts */

extern int                     numCacheFont;

/*-=================================================================|>
||
|> Public functions
||
||==================================================================|>
||
*/
/* round up to next multiple of 4 */
long
pad (const long n)
{
  return ((n + 3) & ~0x3);
}

/* set up the initial definitions in the postscript file */

void
initprocs (struct PrintClient *pc,
           int fd)
{
  XWindowAttributes      *winAttr;
  winfoPtr                winfo;
  Drawable                drawable = pc->win;
  FILE                   *fp = pc->printfp;

  winfo = getwinfo (drawable, winfoTable);
  if (winfo == NULL)
  {
    fprintf (stderr, "couldn't find window id %ld\n", drawable);
    return;
  }
  winAttr = &winfo->attr;

/* postscript has images reverse from real world */

  fprintf (fp, "%%!PS-Adobe-2.0\n");
  fprintf (fp, "%%%%Creator: X to PostScript server\n");
  fprintf (fp, "%%%%Pages: 1\n");
  if (winAttr->width < winAttr->height)
  {
    fprintf (fp, "%%%%Orientation: Portrait\n");
  }
  else
  {
    fprintf (fp, "%%%%Orientation: Landscape\n");
  }
  fprintf (fp, "%%%%EndComments\n");
  if ( pc->prologFile != NULL) 
  {
    char prologStr[ 4096];
    FILE *fp;
    int	n;

    if (( fp = fopen (pc->prologFile, "r")) == NULL)
    {
      fprintf (stderr, "failed to read prolog file %s\n", pc->prologFile);
    }
    else
    {
      n = fread (prologStr, 1, sizeof( prologStr), fp);
      prologStr[ n] = '\0';
      fclose( fp);
      fprintf( pc->printfp, prologStr, pc->pageWidth, pc->pageHeight);
    }
  }


/* definition for image drawing function */

  fprintf (fp, "/imagedraw\n");
  fprintf (fp, "{\n");
  fprintf (fp, "/matrix exch cvlit def\n");
  fprintf (fp, "/depth exch cvlit def\n");
  fprintf (fp, "/height exch cvlit def\n");
  fprintf (fp, "/width exch cvlit def\n");
  fprintf (fp, "/picstr width string def\n");
  fprintf (fp, "width height depth matrix\n");
  fprintf (fp, "    { currentfile picstr readhexstring pop }\n");
  fprintf (fp, "    false 3 colorimage\n");
  fprintf (fp, "} def\n");

/* definition of arc drawing function */

  fprintf (fp, "/elliparc\n");
  fprintf (fp, "{\n");
  fprintf (fp, "/angle2 exch cvlit def\n");
  fprintf (fp, "/angle1 exch cvlit def\n");
  fprintf (fp, "/height exch cvlit def\n");
  fprintf (fp, "/width exch cvlit def\n");
  fprintf (fp, "/y exch cvlit def\n");
  fprintf (fp, "/x exch cvlit def\n");
  fprintf (fp, "1 height width div scale\n");
  fprintf (fp, "x width 2 div add\n");  /* x: center */
  fprintf (fp, "y height 2 div add\n");   /* y: center */
  fprintf (fp, "width 2 div\n");  /* radius */
  fprintf (fp, "angle1 angle2 arc\n");  /* ang1 ang2 arc */
  fprintf (fp, "} def\n");

/* use the window and postscript output dimensions to determine the scaling
   to use for the drawing to this window depending on which size is larger, 
   rotate by 90 deg for landscape */

  if (!pc->landscape)
  {
    fprintf (fp, "[ %d %d] /pagesize exch cvlit def\n",
             pc->pageWidth, pc->pageHeight);
    fprintf (fp, "0 pagesize 1 get translate\n");
    fprintf (fp, "pagesize 0 get %d div ", winAttr->width);
    fprintf (fp, "pagesize 1 get %d div neg ", winAttr->height);
    fprintf (fp, "scale\n");
  }
  else
  {
    fprintf (fp, "4 dict begin\n");
    fprintf (fp, "0 0 %d %d\n", pc->pageWidth, pc->pageHeight);
    fprintf (fp, "/ury exch def /urx exch def\n");
    fprintf (fp, "/lly exch def /llx exch def\n");
    fprintf (fp, "90 rotate\n");
    fprintf (fp, "llx neg llx urx sub lly sub translate\n");
    fprintf (fp, "end\n");
    fprintf (fp, "[ %d %d] /pagesize exch cvlit def\n",
             pc->pageHeight, pc->pageWidth);
    fprintf (fp, "0 pagesize 1 get translate\n");
    fprintf (fp, "pagesize 0 get %d div ", winAttr->width);
    fprintf (fp, "pagesize 1 get %d div neg ", winAttr->height);
    fprintf (fp, "scale\n");
  }
  fprintf (fp, "0 setlinewidth\n");

/* set a basic default font */

  fprintf (fp, "/Times-Roman findfont 7 scalefont setfont\n");
}

/* set the foreground color according to the pixel and the current colormap */

/* apply a color correction vector to a color */
static void
correctColor (
        XColor * color,
        int ncorr,
        float **corrects)
{
  int             ind;
  double          frac;

  ind = ((ncorr - 1) * color->red) / 65536;
  if (ind < ncorr - 1)
  {
    frac = ((ncorr - 1) * color->red) / 65536.0 - ind;
    color->red = frac * color->red * corrects[ind][0] +
      (1.0 - frac) * color->red * corrects[ind + 1][0];
  }
  ind = ((ncorr - 1) * color->blue) / 65536;
  if (ind < ncorr - 1)
  {
    frac = ((ncorr - 1) * color->blue) / 65536.0 - ind;
    color->blue = frac * color->blue * corrects[ind][1] +
      (1.0 - frac) * color->blue * corrects[ind + 1][1];
  }
  ind = ((ncorr - 1) * color->green) / 65536;
  if (ind < ncorr - 1)
  {
    frac = ((ncorr - 1) * color->green) / 65536.0 - ind;
    color->blue = frac * color->green * corrects[ind][2] +
      (1.0 - frac) * color->green * corrects[ind + 1][2];
  }
}

void
setcolor (FD fd,
          FILE * fp,
          unsigned long pixel,
          Colormap cmap)
{
  double                  red,
                          green,
                          blue;
  cmapinfoPtr             cmapinfo;
  XColor                  corrColor;

  cmapinfo = getcmapinfo (cmap, cmapinfoTable);
  if (cmapinfo == NULL) {
    fprintf (stderr, "failed to find cmap %ld\n", cmap);
    return;
  }

/* apply color correction if there's a color correction file */

  if (PCS[fd].correctFile != NULL) {
    if ( cmapinfo->visualID == PSEUDOCOLOR_VISUAL) {
      corrColor.red = cmapinfo->colors[pixel].red;
      corrColor.green = cmapinfo->colors[pixel].green;
      corrColor.blue = cmapinfo->colors[pixel].blue;
    } else {
      unsigned char *vec = ( unsigned char *) &pixel;
      corrColor.blue = vec[ 0] << 8;
      corrColor.green = vec[ 1] << 8;
      corrColor.red = vec[ 2] << 8;
    }
    correctColor (&corrColor, PCS[fd].numCorrect, PCS[fd].corrects);
    red = (double) corrColor.red / 65536.0;
    green = (double) corrColor.green / 65535.0;
    blue = (double) corrColor.blue / 65535.0;
  } else {
    if ( cmapinfo->visualID == PSEUDOCOLOR_VISUAL) {
      red = (double) cmapinfo->colors[pixel].red / 65536.0;
      green = (double) cmapinfo->colors[pixel].green / 65535.0;
      blue = (double) cmapinfo->colors[pixel].blue / 65535.0;
    } else {
      unsigned char *vec = ( unsigned char *) &pixel;	/* format is 0BGR */
      blue = vec[ 1] / 255.0;
      green = vec[ 2] / 255.0;
      red = vec[ 3] / 255.0;
    }
  }
  fprintf (fp, "%.3f %.3f %.3f setrgbcolor\n", red, green, blue);
}

/* set the line width of the current line */

void
setlinewidth (FILE * fp,
              int line_width)
{
  fprintf (fp, "%d setlinewidth\n", line_width);
}

/* set the line cap parameter for the graphics state */

void
setlinecap (FILE * fp,
            int cap_style)
{
  int                     style = 0;

  switch (cap_style) {
  case CapNotLast:
  case CapButt:
    style = 0;
    break;
  case CapRound:
    style = 1;
    break;
  case CapProjecting:
    style = 2;
    break;
  default:
    break;
  }
  fprintf (fp, "%d setlinecap\n", style);
}

/* set the line join style */

void
setlinejoin (FILE * fp,
             int join_style)
{
  int                     style = 0;

  switch (join_style)
  {
  case JoinMiter:
    style = 0;
    break;
  case JoinRound:
    style = 1;
    break;
  case JoinBevel:
    style = 2;
    break;
  default:
    break;
  }
  fprintf (fp, "%d setlinejoin\n", style);
}

/* set the line dash style */

void
setlinedash( FILE 	*fp,
	     gcinfoPtr	gcinfo)
{
    XGCValues	*gcvp = &gcinfo->attr;
    int		ii;

    if (( gcvp->line_style == LineOnOffDash) || 
	( gcvp->line_style == LineDoubleDash)) {
	fprintf( fp, "[ ");
	for ( ii = 0; ii < gcinfo->numDashes; ii++) {
	    fprintf( fp, "%d ", gcinfo->dashes[ ii]);
	}
	fprintf( fp, " ] %d setdash\n", gcvp->dash_offset);
    } else if ( gcvp->line_style == LineSolid) {
	fprintf( fp, "[] 0 setdash\n");
    } else {
	fprintf( stderr, "unknown line style %d\n", gcvp->line_style);
    }
}

/* set an appropriate postscript font from an X bitmap font */

int
setfont (FD fd,
         FILE * fp,
         Font fontId)           /* unsigned long font ID */
{
  int                     i,
                          j,
                          size;
  fontPtr                 font;

/* first, look in the font cache to see if there's a matching font */

  for (i = 0, font = fontTable->fontTable;
       i < fontTable->numFonts;
       i++, font++)
  {
    for (j = 0; j < font->nfid; j++)
    {
      if (fontId == font->fid[j])
      {
        if ((font->postName != NULL) && (font->fontInfo != NULL))
        {
          size = PCS[fd].fontScale * font->fontInfo->ascent;
          fprintf (fp, "/%s findfont %d scalefont setfont\n",
                   font->postName, size);
        }
        return (TRUE);
      }
    }
  }
  fprintf (stderr, "server font id %ld not found\n", fontId);
  return (FALSE);
}

/* fill a list of rectangles */

void
polyfillrectangle (FILE * fp,
                   Window win,
                   XGCValues *gcv,
                   int nrect,
                   XRectangle *rectangles)
{
  while (nrect-- > 0)
  {
    fprintf (fp, "newpath\n");
    fprintf (fp, "%d %d moveto\n",
             (int) rectangles->x, (int) rectangles->y);
    fprintf (fp, "0 %d rlineto\n", (int) rectangles->height);
    fprintf (fp, "%d 0 rlineto\n", (int) rectangles->width);
    fprintf (fp, "0 %d rlineto\n", (int) -rectangles->height);
    fprintf (fp, "closepath fill\n");   /* closepath does last vertex */

/* next rectangle */

    rectangles++;
  }
}

/* fill a list of rectangles */

void
polyrectangle (FILE * fp,
               Window win,
               XGCValues *gcv,
               int nrect,
               XRectangle *rectangles)
{
  while (nrect-- > 0)
  {
    fprintf (fp, "newpath\n");
    fprintf (fp, "%d %d moveto\n",
             (int) rectangles->x, (int) rectangles->y);
    fprintf (fp, "0 %d rlineto\n", (int) rectangles->height);
    fprintf (fp, "%d 0 rlineto\n", (int) rectangles->width);
    fprintf (fp, "0 %d rlineto\n", (int) -rectangles->height);
    fprintf (fp, "closepath stroke\n");   /* closepath does last vertex */

/* next rectangle */

    rectangles++;
  }
}

/* draw a filled polygon */

void
fillpoly (FILE * fp,
          Window win,
          XGCValues *gcv,
          int npoint,
          int coordmode,
          XPoint *points)
{
  fprintf (fp, "newpath\n");
  fprintf (fp, "%d %d moveto\n", (int) points->x, (int) points->y);
  points++;
  while (--npoint > 1)
  {
    if (coordmode == CoordModePrevious)
    {
      fprintf (fp, "%d %d rlineto\n", (int) points->x, (int) points->y);
    }
    else
    {
      fprintf (fp, "%d %d lineto\n", (int) points->x, (int) points->y);
    }
    points++;
  }
  fprintf (fp, "closepath fill\n");
}

/* draw a polyline */

void
polyline (FILE * fp,
          Window win,
          XGCValues *gcv,
          int npoint,
          int coordmode,
          XPoint *points)
{
  fprintf (fp, "newpath\n");
  fprintf (fp, "%d %d moveto\n", (int) points->x, (int) points->y);
  points++;
  while (--npoint > 0)
  {
    if (coordmode == CoordModePrevious)
    {
      fprintf (fp, "%d %d rlineto\n", (int) points->x, (int) points->y);
    }
    else
    {
      fprintf (fp, "%d %d lineto\n", (int) points->x, (int) points->y);
    }
    points++;
  }
  fprintf (fp, "stroke\n");
}

/* draw a polypoint */

void
polypoint (FILE * fp,
           Window win,
           XGCValues *gcv,
           int npoint,
           int coordmode,
           XPoint *points)
{
  fprintf (fp, "newpath\n");
  if (coordmode == CoordModePrevious)
  {
    fprintf (fp, "%d %d moveto\n", (int) points->x, (int) points->y);
    npoint--;
    points++;
  }

/* draw the points as very, very short lines */

  while (npoint-- > 0)
  {
    if (coordmode == CoordModeOrigin)
    {
      fprintf (fp, "%d %d moveto\n", (int) points->x, (int) points->y);
      fprintf (fp, "1 1 rlineto\n");
    }
    else
    {
      fprintf (fp, "%d %d rmoveto\n", (int) points->x, (int) points->y);
      fprintf (fp, "1 1 rlineto\n");
    }
    points++;
  }
  fprintf (fp, "stroke\n");
}

/* arc draw function */

void
polyarc (FILE * fp,
         Window win,
         XGCValues *gcv,
         int narc,
         xArc * arcs)
{
  double                  angle1,
                          angle2;

  fprintf (fp, "newpath\n");
  while (narc--)
  {

/* X arcs are relative, PostScript Arcs are not */

    angle1 = ((double) arcs->angle1) / 64.0;
    angle2 = ((double) arcs->angle2) / 64.0 + angle1;
    fprintf (fp, "%d %d %d %d %f %f elliparc\n",
             arcs->x, arcs->y, arcs->width, arcs->height, -angle2, -angle1);
  }
  fprintf (fp, "stroke\n");
}

/* draw a polysegment set */

void
polysegment (FILE * fp,
             Window win,
             XGCValues *gcv,
             int nseg,
             XSegment *segments)
{
  fprintf (fp, "newpath\n");
  while (nseg-- > 0)
  {
    fprintf (fp, "%d %d moveto\n", segments->x1, segments->y1);
    fprintf (fp, "%d %d lineto\n", segments->x2, segments->y2);
    segments++;
  }
  fprintf (fp, "stroke\n");
}

/* draw text, which is set up in chunks, and may be interspersed with
   font information */

void
polytext8 (FILE * fp,
           Window win,
           XGCValues *gcv,
           int x,
           int y,
           int textlen,
           unsigned char *buffer)
{
  char                    textbuf[1024];
  int                     size,
                          delta;

  fprintf (fp, "%d %d moveto\n", x, y);
  fprintf (fp, "gsave\n");
  fprintf (fp, "1 -1 scale\n");
  while (textlen > 0)
  {

/* we'll skip over the font information for the moment */

    if (buffer[0] == 255)
    {
      buffer += 5;
    }
    else
    {
      size = *buffer++;
      textlen -= size;
      delta = *buffer++;
      fprintf (fp, "0 %d rmoveto\n", delta);
      strncpy (textbuf, (char *) buffer, size);
      textbuf[size] = '\0';
      fprintf (fp, "(%s) show\n", insertQuoteChar( textbuf, "()\\", '\\'));
      buffer += size;
      break;
    }
  }
  fprintf (fp, "grestore\n");
}

/* draw text, which is set up in chunks, and may be interspersed with
   font information */

void
polytext16 (FILE * fp,
            Window win,
            XGCValues *gcv,
            int x,
            int y,
            int ntext,
            unsigned char *buffer)
{

/* stubbed out, because I'm not sure how to do this */

#if 0
  char                    textbuf[512];
  int                     size;

  fprintf (fp, "%d %d moveto\n", x, y);
  while (ntext-- > 0)
  {

/* we'll skip over the font information for the moment */

    if (buffer[0] == 255)
    {
      buffer += 5;
    }
    else
    {
      size = *buffer++;
      strncpy (textbuf, buffer, size);
      fprintf (fp, "(%s) show\n", insertQuoteChar( textbuf, "()\\", '\\'));
      buffer += size;
    }
  }
#endif
}

/* draw text, which is set up in chunks, and may be interspersed with
   font information */

void
imagetext8 (FILE * fp,
            Window win,
            XGCValues *gcv,
            int x,
            int y,
            int textlen,
            unsigned char *buffer)
{
  char                    textbuf[512];
  int                     size,
                          delta;

  fprintf (fp, "%d %d moveto\n", x, y);
  fprintf (fp, "gsave\n");
  fprintf (fp, "1 -1 scale\n");
  while (textlen-- > 0)
  {

/* we'll skip over the font information for the moment */

    if (buffer[0] == 255)
    {
      buffer += 5;
    }
    else
    {
      size = *buffer++;
      textlen -= size;
      delta = *buffer++;
      fprintf (fp, "0 %d rmoveto\n", delta);
      strncpy (textbuf, (char *) buffer, size);
      textbuf[size] = '\0';
      fprintf (fp, "(%s) show\n", insertQuoteChar( textbuf, "()\\", '\\'));
      buffer += size;
      break;
    }
  }
  fprintf (fp, "grestore\n");
}

/* draw text, which is set up in chunks, and may be interspersed with
   font information */

void
imagetext16 (FILE * fp,
             Window win,
             XGCValues *gcv,
             int x,
             int y,
             int ntext,
             unsigned char *buffer)
{

/* stubbed out, because I'm not sure how to do this */

#if 0
  char                    textbuf[512];
  int                     size;

  fprintf (fp, "%d %d moveto\n", x, y);
  while (ntext-- > 0)
  {

/* we'll skip over the font information for the moment */

    if (buffer[0] == 255)
    {
      buffer += 5;
    }
    else
    {
      size = *buffer++;
      strncpy (textbuf, buffer, size);
      fprintf (fp, "(%s) show\n", insertQuoteChar( textbuf, "()\\", '\\'));
      buffer += size;
    }
  }
#endif
}

/* clear an area of a window ( draw a rectangle in the window's background 
   color).  This should really use the window background, but I can't get
   that value from the server ( can you believe it?) */

void
cleararea (FILE * fp,
           Window win,
           XWindowAttributes *attr,
           int x,
           int y,
           int width,
           int height)
{
/* actually, we don't want to do anything */

#if 0
  setcolor (fp, WhitePixel (THE_DISPLAY, DefaultScreen (THE_DISPLAY)),
            attr->colormap);
  fprintf (fp, "%d %d moveto\n", x, y);
  fprintf (fp, "%d 0 rlineto\n", width);
  fprintf (fp, "0 %d rlineto\n", height);
  fprintf (fp, "%d 0 rlineto\n", -width);
  fprintf (fp, "closepath fill\n");
#endif
}

/* put an image */

void
putimage (int fd,
          FILE * fp,
          Window win,
          XGCValues *gcv,
          int depth,
          int width,
          int height,
          int x,
          int y,
          int leftpad,
          int format,
          int numColors,
          XColor *colors,
          unsigned char *data)
{
    XColor                 *corrColors;
    int                     i;

/* even though it's not documented anywhere, it seems that the width
   needs to be padded by 2 */

    fprintf (fp, "gsave\n");

  /* if there's a color correction file, apply it to the colors in the image */

    if ( depth == 8) {    	
	if (PCS[fd].correctFile != NULL) {
	    corrColors = (XColor *) malloc (sizeof (XColor) * numColors);
	    for (i = 0; i < numColors; i++) {
		corrColors[i] = colors[i];
		correctColor (&corrColors[i], PCS[fd].numCorrect,
			      PCS[fd].corrects);
	    }
	    printImageGrey ((char *) data, x, y, pad (width), height, depth,
			    corrColors, numColors, fp,
			    PRINT_OPTIONS.indexCspaceSupport);
	    free (corrColors);
	} else {
	    printImageGrey ((char *) data, x, y, pad (width), height, depth,
			    colors, numColors, fp,
			    PRINT_OPTIONS.indexCspaceSupport);
	}
    } else if ( depth == 24) {
	printImage24((char *) data, x, y, width, height, depth, fp);
    }
    fprintf (fp, "grestore\n");
}

/* if the current GC pointer changes, reset it and set a new graphics state */

void
setCurrGc (FD 		fd,
           FILE 	*fp,
	   gcinfoPtr	gcinfo,
           Colormap 	cmap)
{
  XGCValues *gcvp = &gcinfo->attr;
   
  if (gcvp != currGcvp) {

/* plane mask, background and function haven't been addressed yet */

    setcolor (fd, fp, gcvp->foreground, cmap);
    setlinewidth (fp, gcvp->line_width);
    setlinecap (fp, gcvp->cap_style);
    setlinejoin (fp, gcvp->join_style);
    setlinedash (fp, gcinfo);
    setfont (fd, fp, gcvp->font);
    currGcvp = gcvp;
  }
}

/* set the values in a GC given the flags and the data buffer from a request,
   and if it's the current GC, set the value immediately */

void
setGcvalues (FD fd,
             FILE * fp,
             XGCValues *gcvp,
             unsigned long value_mask,
             unsigned long *request,  /* each val is a long */
             Colormap cmap)
{
  if (value_mask & GCFunction)
  {
    gcvp->function = *request++;
  }
  if (value_mask & GCPlaneMask)
  {
    gcvp->plane_mask = *request++;
  }

/* set the foreground color */

  if (value_mask & GCForeground)
  {
    gcvp->foreground = *request++;
    if (gcvp == currGcvp)
    {
      setcolor (fd, fp, gcvp->foreground, cmap);
    }
  }
  if (value_mask & GCBackground)
  {
    gcvp->background = *request++;
  }
  if (value_mask & GCLineWidth)
  {
    gcvp->line_width = *request++;
  }
  if (value_mask & GCLineStyle)
  {
    gcvp->line_style = *request++;
  }
  if (value_mask & GCCapStyle)
  {
    gcvp->cap_style = *request++;
  }
  if (value_mask & GCJoinStyle)
  {
    gcvp->join_style = *request++;
  }
  if (value_mask & GCFillStyle)
  {
    gcvp->fill_style = *request++;
  }
  if (value_mask & GCFillRule)
  {
    gcvp->fill_rule = *request++;
  }
  if (value_mask & GCTile)
  {
    gcvp->tile = *request++;
  }
  if (value_mask & GCStipple)
  {
    gcvp->stipple = *request++;
  }
  if (value_mask & GCTileStipXOrigin)
  {
    gcvp->ts_x_origin = *request++;
  }
  if (value_mask & GCTileStipYOrigin)
  {
    gcvp->ts_y_origin = *request++;
  }
  if (value_mask & GCFont)
  {
    gcvp->font = *request++;
    setfont (fd, fp, gcvp->font);
  }
  if (value_mask & GCSubwindowMode)
  {
    gcvp->subwindow_mode = *request++;
  }
  if (value_mask & GCGraphicsExposures)
  {
    gcvp->graphics_exposures = *request++;
  }
  if (value_mask & GCClipXOrigin)
  {
    gcvp->clip_x_origin = *request++;
  }
  if (value_mask & GCClipYOrigin)
  {
    gcvp->clip_y_origin = *request++;
  }
  if (value_mask & GCClipMask)
  {
    gcvp->clip_mask = *request++;
  }
  if (value_mask & GCDashOffset)
  {
    gcvp->dash_offset = *request++;
  }
  if (value_mask & GCDashList)
  {
    gcvp->dashes = *request++;
  }
  if (value_mask & GCArcMode)
  {
    gcvp->arc_mode = *request++;
  }

/* if it's the current GC, then set the appropriate attributes */

  if (gcvp == currGcvp) {
    setcolor (fd, fp, gcvp->foreground, cmap);
    setlinewidth (fp, gcvp->line_width);
    setlinecap (fp, gcvp->cap_style);
    setlinejoin (fp, gcvp->join_style);
    setfont (fd, fp, gcvp->font);
  }
}

/* set the values in a GC given the flags and the data buffer from a request,
   and if it's the current GC, set the value immediately */

void
copyGcvalues (XGCValues *gcvp,
              XGCValues *srcGcvp,
              unsigned long value_mask)
{
  if (value_mask & GCFunction)
  {
    gcvp->function = srcGcvp->function;
  }
  if (value_mask & GCPlaneMask)
  {
    gcvp->plane_mask = srcGcvp->plane_mask;
  }

/* set the foreground color */

  if (value_mask & GCForeground)
  {
    gcvp->foreground = srcGcvp->foreground;
  }
  if (value_mask & GCBackground)
  {
    gcvp->background = srcGcvp->background;
  }
  if (value_mask & GCLineWidth)
  {
    gcvp->line_width = srcGcvp->line_width;
  }
  if (value_mask & GCLineStyle)
  {
    gcvp->line_style = srcGcvp->line_style;
  }
  if (value_mask & GCCapStyle)
  {
    gcvp->cap_style = srcGcvp->cap_style;
  }
  if (value_mask & GCJoinStyle)
  {
    gcvp->join_style = srcGcvp->join_style;
  }
  if (value_mask & GCFillStyle)
  {
    gcvp->fill_style = srcGcvp->fill_style;
  }
  if (value_mask & GCFillRule)
  {
    gcvp->fill_rule = srcGcvp->fill_rule;
  }
  if (value_mask & GCTile)
  {
    gcvp->tile = srcGcvp->tile;
  }
  if (value_mask & GCStipple)
  {
    gcvp->stipple = srcGcvp->stipple;
  }
  if (value_mask & GCTileStipXOrigin)
  {
    gcvp->ts_x_origin = srcGcvp->ts_x_origin;
  }
  if (value_mask & GCTileStipYOrigin)
  {
    gcvp->ts_y_origin = srcGcvp->ts_y_origin;
  }
  if (value_mask & GCFont)
  {
    gcvp->font = srcGcvp->font;
  }
  if (value_mask & GCSubwindowMode)
  {
    gcvp->subwindow_mode = srcGcvp->subwindow_mode;
  }
  if (value_mask & GCGraphicsExposures)
  {
    gcvp->graphics_exposures = srcGcvp->graphics_exposures;
  }
  if (value_mask & GCClipXOrigin)
  {
    gcvp->clip_x_origin = srcGcvp->clip_x_origin;
  }
  if (value_mask & GCClipYOrigin)
  {
    gcvp->clip_y_origin = srcGcvp->clip_y_origin;
  }
  if (value_mask & GCClipMask)
  {
    gcvp->clip_mask = srcGcvp->clip_mask;
  }
  if (value_mask & GCDashOffset)
  {
    gcvp->dash_offset = srcGcvp->dash_offset;
  }
  if (value_mask & GCDashList)
  {
    gcvp->dashes = srcGcvp->dashes;
  }
  if (value_mask & GCArcMode)
  {
    gcvp->arc_mode = srcGcvp->arc_mode;
  }
}

/* copy the pixmap from the screen and draw it on the page */

void
copyarea (FILE * fp,
          Drawable src,
          Window win,
          int src_x,
          int src_y,
          int dstX,
          int dstY,
          int width,
          int height,
          int numColors,
          XColor *colors)
{
  XImage                 *image;
  int                     screen = DefaultScreen (THE_DISPLAY);
  winfoPtr                winfo = getwinfo (src, winfoTable);

  if (!winfo)
  {
/* use an XImage and GetImage to get the area being copied */
    image = XCreateImage (THE_DISPLAY,
                          DefaultVisual (THE_DISPLAY, screen),
                          DefaultDepth (THE_DISPLAY, screen),
                          ZPixmap, 0, NULL, width, height, 8, 0);
    image->data = (char *) malloc (image->bytes_per_line * image->height);
    XGetImage (THE_DISPLAY, src, src_x, src_y, width, height, 0, ZPixmap);
    printImageGrey (image->data, dstX, dstY,
                    pad (width), height,
                    DefaultDepth (THE_DISPLAY, screen), colors,
                    numColors, fp, PRINT_OPTIONS.indexCspaceSupport);
    free (image->data);
    free (image);
  }
  else
  {

/*  I can't seem to find an operator for copying areas in PostScript */

  }
}

/* match a string with '*'s and '?'s */

static char            *
strqncmp (char *str1,
          char *str2,
          int n)
{
  int                     i;

  for (i = 0; i < n; i++)
  {
    if (*str2 == '*')
    {
      return (str2 + 1);
    }
    if (*str1 != *str2)
    {
      if (*str1 != '?')
      {
        return (NULL);
      }
    }
    str1++;
    str2++;
  }
  return (str2);
}

/* see if two font wildcards match */

static int
matchFontWild (
                char *wild1,
                char *wild2)

{
  char                   *p, *q, *r = NULL;

  for (p = strtok (wild1, "*"); p != NULL; p = strtok (NULL, "*")) {
    q = wild2; 
    r = strqncmp (p, q, strlen (p));
    if (!r) {
      return FALSE;
    }
  }
  return (*r == '\0');
}

char                   *
getPostName (char *xfont)
{
  char                  **p;

/* names are x font names interspersed with postscript names */

  for (p = fontlist; *p; p += 2)
    if (matchFontWild (xfont, *p))
      return (p[1]);

  return 0;
}

/* initialize the default values for the print clients */

void initClient( struct PrintClient *pcs)
{
    pcs->preserveAspect = 0;
    pcs->fontScale = 1;
    pcs->pixmapScale = 1;
    pcs->pageWidth = 612;  /* 8 1/2 x 11 is default size */
    pcs->pageHeight = 792;
    pcs->printDirect = FALSE;
    pcs->landscape = FALSE;
    pcs->sequence = 0;
    pcs->printfp = NULL;
    pcs->printDirect = FALSE;
}

void
initPrintClients (
                   struct PrintClient pcs[],
                   const int numclients)

{
  int                     ii = 0;

  for (; ii < numclients; ii++) {
    initClient( &pcs[ ii]);
  }
}

/* in a string, all characters which are in the set of characters to
   be quoted have to quote character inserted before them. */

char *insertQuoteChar( char 	*str, 
		       char 	*charsToQuote,
		       int	quoteChar)
{
    static char resStr[ 512];
    char	*p;

    p = resStr;
    while ( *str != '\0') {
	if ( strchr( charsToQuote, *str) != NULL) {
	    *p++ = quoteChar;
	}
	*p++ = *str++;
    }
    *p = '\0';
    return( resStr);
}
